#include <iostream.h>
#include <assert.h>
#include "BST.H"

template <class T> 
BSTNode<T>::BSTNode(T v, BSTNode<T> *l, BSTNode<T> *r) 
: value(v), left(l), right(r) 
{
}

template <class T>
int 
BSTNode<T>::isEmpty(BSTNode<T> *n) 
{ 
	return n == NULL; 
}
	
template <class T>
int 
BSTNode<T>::isLeaf(BSTNode<T> * n) 
{ 
	return isEmpty(n->left) && isEmpty(n->right); 
}

template <class T>
void 
BSTNode<T>::insert(T x) 
{
	if (x == value)
		return;
	else if (x < value)
		if (!isEmpty(left))
			left->insert(x);
		else
			left = new BSTNode(x, NULL, NULL);
	else
		if (!isEmpty(right))
			right->insert(x);
		else
			right = new BSTNode(x, NULL, NULL);
}


template <class T>
int 
BSTNode<T>::find(T x) 
{
	if (x == value) 
		return 1;
	else if (x < value)
		if (isEmpty(left))
			return 0;
		else
			return left->find(x);
	else
                if (isEmpty(right))
                	return 0;
		else
			return right->find(x);
}

template <class T>
BSTNode<T> * 
BSTNode<T>::remove(T x) 
{
	BSTNode<T> * tmp;

	if (x < value) {
		if (isEmpty(left))
			return this;

		tmp = left->remove(x);
		if (tmp == NULL) {
			delete left;
			left = tmp;
		} 
		assert(tmp == left);
		
		return this;
	}

	if (x > value) {
		if (isEmpty(right))
			return this;

		tmp = right->remove(x);
		if (tmp == NULL) {
			delete right;
			right = tmp;
		}
		assert(right == tmp);
	
		return this;
	}

	if (isLeaf(this))	
		return NULL;

	//  Now the fun begins...
	if (isEmpty(left)) {
		value = right->findmin();
		tmp = right->remove(value);

		if (tmp == NULL) {
			delete right;
			right = tmp;
		}
		assert(tmp == right);

		return this;
	} 

	value = left->findmax();
	tmp = left->remove(value);

	if (tmp == NULL) {
		delete left;
		left = tmp;
	}
	assert(tmp == left);

	return this;
}


template <class T>
T 
BSTNode<T>::findmin() 
{
	if (isEmpty(left)) 
		return value;	
	return left->findmin();
}	


template <class T>
T 
BSTNode<T>::findmax() 
{ 
	if (isEmpty(right))
		return value;
	return right->findmax();
}

	
//  ----  BST Code below this line ----
template <class T>
BST<T>::BST() : root(NULL) 
{
}


template <class T>
BST<T>::BST(BSTNode<T> * r) : root(r)
{
}

	
template <class T>
void 
BST<T>::insert(T x) 
{ 
	if (root == NULL) 
		root = new BSTNode<T>(x, NULL, NULL);
	root->insert(x); 
}

template <class T>
void 
BST<T>::remove(T x) 
{ 
	if (root == NULL)
		return;

	root->remove(x); 
}

template <class T>
int 
BST<T>::find(T x) 
{  
	if (root == NULL) 
		return 0;

	return root->find(x);
}

template <class T>
inline int 
BST<T>::isEmpty() 
{ 
	return root == NULL; 
}

template <class T>
inline T 
BST<T>::value() 
{ 
	return root->value; 
}

template <class T>
BST<T>
BST<T>::left() 
{ 
	return BST(root->left); 
}

template <class T>
BST<T> 
BST<T>::right() 
{ 
	return BST(root->right); 
}
	
template <class T>
ostream & operator <<(ostream & os, BST<T> t) 
{
	if (t.isEmpty()) 
		return os << "empty";
	
	return os << "(" << t.value() << "," 
		  << t.left() << "," 
		  << t.right() << ")";
}


